﻿using QPR.QPRSuite.NCalc;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;

namespace QPR.QPRSuite.AcceleratorTools {
    public class ContextManager : IContextManager {

        protected ILogger logger = null;
        protected IWebServiceClient webServiceClient = null;
        protected IModellingClient modellingClient = null;
        protected IMetricsClient metricsClient = null;
        protected ExpressionEngine expressionEngine = null;
        protected Dictionary<string, string> applicationParameters = new Dictionary<string, string>();
        private static string WEB_SERVICE_ENDPOINT_NAME = "WSHttpBinding_IService";
        public static string EXPRESSION_LANGUAGE_VERSION = "1.6.0 (1.1.2016)";
        private string[,] DEFAULT_PARAMETERS = new string[,] {
          { "applicationversion", "not defined" },
          { "wcfsecuritymode", "none" },
          { "qprauthenticationmode", "password" },
          { "solutionspecificcustomfunctionsnamespace", "" },
          { "username", "" },
          { "password", "" },
          { "webservicesessionid", ""},
          { "qprwebserviceaddress", ""}
        };
        private static string ILLEGAL_CHARACTERS_IN_FILENAMES = "/\\?*:|\"<>";

        public ContextManager(Dictionary<string, string> applicationParameters) {
            this.applicationParameters = applicationParameters;
            for (int i = 0; i < DEFAULT_PARAMETERS.Length / 2; i++) {
                if (!this.applicationParameters.ContainsKey(DEFAULT_PARAMETERS[i, 0])) this.applicationParameters.Add(DEFAULT_PARAMETERS[i, 0], DEFAULT_PARAMETERS[i, 1]);
            }
            if (!(new String[] { "none", "message", "transport", "transportwithmessagecredential", "notconfigured" }).Contains(this.getMandatoryApplicationSetting("wcfsecuritymode").ToLower())) throw new Exception("wcfsecuritymode is not configured properly. It must be one of following: none, message, transport, transportwithmessagecredential, notconfigured. Found value is " + this.getMandatoryApplicationSetting("wcfsecuritymode"));
            WCFSecurityMode WCFSecurityMode;
            switch (this.getMandatoryApplicationSetting("wcfsecuritymode").ToLower()) {
                case "none":
                    WCFSecurityMode = WCFSecurityMode.none;
                    break;
                case "message":
                    WCFSecurityMode = WCFSecurityMode.message;
                    break;
                case "transport":
                    WCFSecurityMode = WCFSecurityMode.transport;
                    break;
                case "transportwithmessagecredential":
                    WCFSecurityMode = WCFSecurityMode.transportwithmessagecredential;
                    break;
                default:
                    WCFSecurityMode = WCFSecurityMode.notconfigured;
                    break;
            }

            if (!(new String[] { "password", "windows", "passedsession", "fixedcredentials" }).Contains(this.getMandatoryApplicationSetting("qprauthenticationmode").ToLower())) throw new Exception("wcfsecuritymode is not configured properly. It must be one of following: password, windows, passedsession, fixedcredentials. Found value is " + this.getMandatoryApplicationSetting("qprauthenticationmode"));
            QPRAuthenticationMode QPRAuthenticationMode;
            switch (this.getMandatoryApplicationSetting("qprauthenticationmode").ToLower()) {
                case "password":
                    QPRAuthenticationMode = QPRAuthenticationMode.password;
                    break;
                case "windows":
                    QPRAuthenticationMode = QPRAuthenticationMode.windows;
                    break;
                case "passedsession":
                    QPRAuthenticationMode = QPRAuthenticationMode.passedsession;
                    break;
                default:
                    QPRAuthenticationMode = QPRAuthenticationMode.password;
                    break;
            }
            if (QPRAuthenticationMode == QPRAuthenticationMode.windows && (WCFSecurityMode == WCFSecurityMode.none || WCFSecurityMode == WCFSecurityMode.transport)) throw new Exception("When qprauthenticationmode is \"windows\", wcfsecuritymode must be \"message\".");
            this.webServiceClient = new WebServiceClient(this, WCFSecurityMode, QPRAuthenticationMode, this.getMandatoryApplicationSetting("webservicesessionid"), this.getMandatoryApplicationSetting("qprwebserviceaddress"), this.getMandatoryApplicationSetting("username"), this.getMandatoryApplicationSetting("password"), ContextManager.WEB_SERVICE_ENDPOINT_NAME);
            this.modellingClient = new ModellingClient(this);
            this.metricsClient = new MetricsClient(this);
            this.expressionEngine = new ExpressionEngine(this, this.getMandatoryApplicationSetting("solutionspecificcustomfunctionsnamespace"));
            this.logger = new MemoryLogger(this);
        }

        public ILogger getLogger() {
            return this.logger;
        }

        public IWebServiceClient getWebServiceClient() {
            return this.webServiceClient;
        }

        public ExpressionEngine getExpressionEngine() {
            return this.expressionEngine;
        }

        public IModellingClient getModellingClient() {
            return this.modellingClient;
        }

        public IMetricsClient getMetricsClient() {
            return this.metricsClient;
        }

        public string getMandatoryNotEmptyApplicationSetting(string settingName) {
            string settingValue = getMandatoryApplicationSetting(settingName);
            if (string.IsNullOrEmpty(settingValue)) throw new Exception("Setting \"" + settingName + "\" is empty although it must contain a value.");
            return settingValue;
        }

        public string getMandatoryApplicationSetting(string settingName) {
            if (!this.applicationParameters.ContainsKey(settingName)) throw new Exception("Setting \"" + settingName + "\" is not found. This is probably an error in the application configuration.");
            return this.applicationParameters[settingName];
        }

        public string getOptionalApplicationSetting(string settingName) {
            if (this.applicationParameters.ContainsKey(settingName)) return this.applicationParameters[settingName];
            return null;
        }

        public void setApplicationSetting(string name, string value) {
            this.applicationParameters[name] = value;
        }

        public string getApplicationVersion() {
            return this.applicationParameters["applicationversion"];
        }

        public string getExpressionLanguageVersion() {
            return ContextManager.EXPRESSION_LANGUAGE_VERSION;
        }

        public string removeInvalidFileNameChars(string input) {
            if (input == null) return null;
            string output = "";
            for (int i = 0; i < input.Length; i++) {
                if (!ContextManager.ILLEGAL_CHARACTERS_IN_FILENAMES.Contains(Convert.ToString(input[i]))) {
                    output = output + input[i];
                }
            }
            return output;
        }

        public string removeInvalidXMLChars(string input) {
            return Regex.Replace(input, @"[\u0000-\u0008\u000B\u000C\u000E-\u001F]", "");
        }

        public void Dispose() {
            this.webServiceClient.Dispose();
            this.modellingClient.Dispose();
            this.metricsClient.Dispose();
            this.logger.Dispose();
        }

    }
}
