﻿using NCalc;
using QPR.QPRSuite.AcceleratorTools;
using System;
using System.Collections.Generic;

namespace QPR.QPRSuite.NCalc {
    public class buildhierarchy : CustomFunctionBase {

        private static string ROOT_ID = "root";

        private Dictionary<string, List<string>> childObjects = new Dictionary<string, List<string>>();
        private Dictionary<string, int> rowNumbers = new Dictionary<string, int>();
        private Dictionary<string, int> numberOfInstancesAdded = new Dictionary<string, int>();
        private Dictionary<string, string> correspondingObjectIds = new Dictionary<string, string>();
        private DataSet inputDataSet;
        private DataSet outputDataSet;
        private string filteringFormula = null;
        private string sortingFormula = null;

        public buildhierarchy(FunctionArgs args, IContextManager context)
            : base(args, context, 7, new string[] { "DataSet", "String", "String", "String", "String", "String", "String", "String", "String" }) {
        }

        public override object evaluate() {
            this.inputDataSet = (DataSet)this.parameters[0];
            if (this.inputDataSet == null) return null;
            int instanceIdColumnNumber = this.inputDataSet.getColumnNumber((string)this.parameters[1]);
            if (instanceIdColumnNumber == -1) throw new Exception("Instance id column not found in BuildHierarchy function");
            int objectIdColumnNumber = this.inputDataSet.getColumnNumber((string)this.parameters[2]);
            if (objectIdColumnNumber == -1) throw new Exception("Object id column not found in BuildHierarchy function");
            int parentObjectIdColumnNumber = this.inputDataSet.getColumnNumber((string)this.parameters[3]);
            if (parentObjectIdColumnNumber == -1) throw new Exception("Parent object id column not found in BuildHierarchy function");
            string hierarchyIdColumnName = (string)this.parameters[4];
            if (string.IsNullOrEmpty(hierarchyIdColumnName)) throw new Exception("Hierarchy id column name is empty in BuildHierarchy function");
            if (this.inputDataSet.getColumnNumber(hierarchyIdColumnName) > -1) throw new Exception("Hierarchy id column name was already in the dataset in BuildHierarchy function");
            string hierarchyParentIdColumnName = (string)this.parameters[5];
            if (string.IsNullOrEmpty(hierarchyParentIdColumnName)) throw new Exception("Hierarchy parent id column name is empty in BuildHierarchy function");
            if (this.inputDataSet.getColumnNumber(hierarchyIdColumnName) > -1) throw new Exception("Hierarchy parent id column name was already in the dataset in BuildHierarchy function");
            string levelColumnName = (string)this.parameters[6];
            if (string.IsNullOrEmpty(levelColumnName)) throw new Exception("Level column name is empty in BuildHierarchy function");
            if (this.inputDataSet.getColumnNumber(hierarchyIdColumnName) > -1) throw new Exception("Level column name was already in the dataset in BuildHierarchy function");
            if (this.parameters.Length > 7) this.filteringFormula = (string)this.parameters[7];
            if (this.parameters.Length > 8) this.sortingFormula = (string)this.parameters[8];

            this.childObjects.Add(buildhierarchy.ROOT_ID, new List<string>());
            for (int i = 0; i < this.inputDataSet.rowsCount(); i++) {
                string objectId = (string)this.inputDataSet.getCell(objectIdColumnNumber, i);
                if (!this.childObjects.ContainsKey(objectId)) this.childObjects.Add(objectId, new List<string>());
            }
            for (int i = 0; i < this.inputDataSet.rowsCount(); i++) {
                string instanceId = (string)this.inputDataSet.getCell(instanceIdColumnNumber, i);
                string objectId = (string)this.inputDataSet.getCell(objectIdColumnNumber, i);
                string parentObjectId = (string)this.inputDataSet.getCell(parentObjectIdColumnNumber, i);
                if (!this.childObjects.ContainsKey(parentObjectId)) parentObjectId = buildhierarchy.ROOT_ID;
                this.childObjects[parentObjectId].Add(instanceId);
                this.rowNumbers[instanceId] = i;
                this.numberOfInstancesAdded[instanceId] = 0;
                this.correspondingObjectIds[instanceId] = objectId;
            }
            List<string> outputDatasetColumns = new List<string>(this.inputDataSet.getColumns());
            outputDatasetColumns.Add(hierarchyIdColumnName);
            outputDatasetColumns.Add(hierarchyParentIdColumnName);
            outputDatasetColumns.Add(levelColumnName);
            this.outputDataSet = new DataSet(outputDatasetColumns.ToArray());
            List<string> topLevelNodes = this.childObjects[buildhierarchy.ROOT_ID];
            for (int i = 0; i < topLevelNodes.Count; i++) {
                buildhierarchyNode(topLevelNodes[i], buildhierarchy.ROOT_ID, 0);
            }
            return outputDataSet;
        }

        private void buildhierarchyNode(string instanceId, string hierarchyParentId, int level) {
            List<Object> rowData = new List<Object>(this.inputDataSet.getRow(rowNumbers[instanceId]).getData());
            if (!string.IsNullOrEmpty(this.filteringFormula)) {
                Dictionary<string, object> filterArguments = new Dictionary<string, object>();
                for (int i = 0; i < rowData.Count; i++) {
                    filterArguments[this.inputDataSet.getColumnName(i)] = rowData[i];
                }
                bool filteringResult = this.context.getExpressionEngine().evaluateExpressionAsBoolean(filteringFormula, filterArguments);
                if (!filteringResult) return;
            }

            string hierarchyId = null;
            if (this.numberOfInstancesAdded[instanceId] == 0) {
                hierarchyId = instanceId;
            } else {
                hierarchyId = instanceId + "." + this.numberOfInstancesAdded[instanceId];
            }
            this.numberOfInstancesAdded[instanceId]++;
            rowData.Add(hierarchyId);
            rowData.Add(hierarchyParentId);
            rowData.Add(level);
            this.outputDataSet.addRow(rowData.ToArray());
            string objectId = this.correspondingObjectIds[instanceId];
            if (string.IsNullOrEmpty(objectId)) return;
            if (!this.childObjects.ContainsKey(objectId)) return;
            List<string> childNodes = this.childObjects[this.correspondingObjectIds[instanceId]];
            for (int i = 0; i < childNodes.Count; i++) {
                buildhierarchyNode(childNodes[i], hierarchyId, level + 1);
            }
        }

    }
}