﻿using proguide;
using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;

namespace QPR.QPRSuite.AcceleratorTools {
    public class ModellingClient : IModellingClient {

        private PGApplication client = null;
        private PGModel model = null;
        private IContextManager context = null;
        private int openModelId = 0;
        public static int DEFAULT_PORT = 4251;

        public ModellingClient(IContextManager context) {
            this.context = context;
        }

        private void checkClientOpened() {
            if (this.client == null) {
                this.client = new PGApplication();
                string port = this.context.getOptionalApplicationSetting("modellingserverport");
                if (string.IsNullOrEmpty(port)) port = ModellingClient.DEFAULT_PORT + "";
                int returnCode = this.client.ConnectToServer(this.context.getMandatoryNotEmptyApplicationSetting("servername"), Int32.Parse(port));
                checkReturnCode(returnCode, "connecting to server");
                string uiMode = this.context.getOptionalApplicationSetting("showclient");
                if (string.IsNullOrEmpty(uiMode)) uiMode = "0";
                this.client.SetUIMode(Int32.Parse(uiMode) == 1 ? proguide.UIMode.UI_NORMAL : proguide.UIMode.UI_SILENT);
                returnCode = this.client.Authenticate(this.context.getMandatoryNotEmptyApplicationSetting("username"), this.context.getMandatoryApplicationSetting("password"), this.context.getOptionalApplicationSetting("domain"));
                checkReturnCode(returnCode, "authenticating to server");
            }
        }

        public void openModel(int modelId) {
            this.closeModel(1);
            object outModel;
            int returnCode = this.client.OpenServerModel(modelId, out outModel);
            checkReturnCode(returnCode, "opening server model");
            this.model = (PGModel)outModel;
            this.openModelId = modelId;
        }

        private void closeModel(int save) {
            if (this.model == null) return;
            int returnCode = this.model.CloseModel(save);
            checkReturnCode(returnCode, "closing model");
            this.model = null;
            this.openModelId = 0;
        }

        public void closeModelWithoutSaving() {
            this.closeModel(0);
        }

        private void checkModelOpened(int modelId) {
            this.checkClientOpened();
            if (modelId == 0 && this.openModelId != 0) { //use currently open model
                return;
            } else if (this.openModelId == 0) { //no model opened
                this.openModel(modelId);
                this.openModelId = modelId;
            } else if (this.openModelId != modelId) { //wrong model opened
                this.closeModel(1);
                this.openModel(modelId);
                this.openModelId = modelId;
            }
        }

        public int createElement(int modelId, string typename, string elementName, string parameters) {
            this.checkModelOpened(modelId);
            object elementId = null;
            int returnCode = this.model.CreateElement(typename, elementName, parameters, ref elementId);
            checkReturnCode(returnCode, "creating element (typename=\"" + typename + "\", name=\"" + elementName + "\", parameters=\"" + parameters + "\")");
            return (int)elementId;
        }

        public string createInstance(int modelId, int elementId, string parameters) {
            this.checkModelOpened(modelId);
            int returnCode;
            object instanceId = null;
            returnCode = this.model.CreateInstance(elementId, parameters, ref instanceId);
            checkReturnCode(returnCode, "creating element instance for elementId=\"" + elementId + "\" with parameters \"" + parameters + "\"");
            return (string)instanceId;
        }

        public void setProperty(int modelId, int elementId, string propertyName, object value) {
            this.checkModelOpened(modelId);
            int returnCode;
            returnCode = this.model.SetProperty(elementId, propertyName, value);
            string valueToLog = "";
            if (value is string) {
                valueToLog = (string)value;
            } else if (value is int[,]) {
                int[,] valueAsArray = (int[,])value;
                List<string> valueAsList = new List<string>();
                for (int i = 0; i < (valueAsArray.Length / 3); i++) {
                    valueAsList.Add("(" + valueAsArray[i, 0] + "," + valueAsArray[i, 1] + "," + valueAsArray[i, 2] + ")");
                }
                valueToLog = String.Join(",", valueAsList.ToArray<string>());
            }
            checkReturnCode(returnCode, "setting property (elementId=\"" + elementId + "\", property=\"" + propertyName + "\", value=\"" + valueToLog + "\")");
        }

        public object getProperty(int modelId, int elementId, string propertyName) {
            this.checkModelOpened(modelId);
            int returnCode;
            object value;
            returnCode = this.model.GetProperty(elementId, propertyName, out value);
            checkReturnCode(returnCode, "getting property (elementId=\"" + elementId + "\", property=\"" + propertyName + "\")");
            return value;
        }

        public int deleteElement(int modelId, int elementId, string parameters) {
            this.checkModelOpened(modelId);
            int returnCode;
            returnCode = this.model.DeleteElement(elementId, parameters);
            checkReturnCode(returnCode, "deleting element (elementId=\"" + elementId + "\", parameters=\"" + parameters + "\")");
            return elementId;
        }

        public void saveModel() {
            int returnCode = this.model.SaveModel();
            checkReturnCode(returnCode, "saving model");
        }

        public int getModelIdByName(string modelName) {
            this.checkClientOpened();
            object models;
            int returnCode;
            returnCode = this.client.GetServerModels(out models);
            checkReturnCode(returnCode, "getting server models");
            object[,] modelsAsArray = (object[,])models;
            for (int i = 0; i < modelsAsArray.Length / 14; i++) {
                if (((string)modelsAsArray[i, 9]) == modelName) {
                    return (int)modelsAsArray[i, 1];
                }
            }
            throw new Exception("model with name \"" + modelName + " \" was not found");
        }

        private void checkReturnCode(int returnCode, string task) {
            this.context.getLogger().writeLog((task.ElementAt(0) + "").ToUpper() + task.Substring(1) + " (" + returnCode + ")", LogLevel.verbose);
            if (returnCode != 0) {
                throw new Exception("QPR Modelling Client returned an error code " + returnCode + ". Error occurred when " + task + ". (http://kb.qpr.com/qpr2015-1/index.html?returnvalues.htm)");
            }
        }

        public void Dispose() {
            int returnCode;
            if (this.client != null) {
                returnCode = this.client.Quit(0);
                checkReturnCode(returnCode, "closing client");
            }
            this.model = null;
            this.client = null;
            this.openModelId = 0;
        }

        public static string parametersDictionaryToString(Dictionary<string, string> inputDictionary) {
            List<string> parametersAsList = new List<string>();
            foreach (string parameter in inputDictionary.Keys) {
                parametersAsList.Add(parameter + "=" + inputDictionary[parameter]);
            }
            return String.Join(";", parametersAsList.ToArray<string>());
        }

    }
}
